from .neural import NeuralClassifier

class FeedForward(NeuralClassifier):
    def __init__(self, classes, voc, embeddings, **kwargs):
        super().__init__(classes, voc)
        self.model = self._defaultmodel(embeddings, **kwargs)

    def _defaultmodel(self, embeddings, hidden=None, activation='relu', optimizer='adagrad', softmax=True, loss='categorical_crossentropy'):
        from keras.models import Sequential
        from keras.layers.core import Dense, Dropout

        def embs(dims, mkemb):
            if type(self.voc) == tuple:
                arch = [(v, (None,) + tuple(params)) for v, params in zip(self.voc, dims)]
            elif type(self.voc) == dict:
                arch = [(self.voc[params[0]], params) for params in dims]
            else:
                arch = (self.voc, dims)
                return mkemb(arch)

            from keras.layers import Merge
            parts = [mkemb(x) for x in arch]
            return Sequential([Merge(parts, mode='concat')])

        def initemb(voc, pretrained, trainable=True):
            from keras.layers.embeddings import Embedding
            if type(pretrained) == str:
                with open(pretrained, "rt") as fh:
                    pretrained = {}
                    for token in fh:
                        token = token.split()
                        pretrained[token[0]] = [float(f) for f in token[1:]]
            import numpy as np
            embeddings = [None] * (max(voc.values()) + 1)
            for w, v in pretrained.items():
                if w == "<UNK>": continue
                if w not in voc: embeddings.append(None)
                i = voc.setdefault(w, len(embeddings) - 1)
                embeddings[i] = np.array(v)
            embeddings = [pretrained["<UNK>"] if v is None else v for v in embeddings]
            return Embedding(len(embeddings), embeddings[0].shape[0], weights=[np.stack(embeddings)], trainable=trainable)

        def mkemb(params):
            from keras.layers import InputLayer
            from keras.layers.core import Flatten
            from keras.layers.embeddings import Embedding
            voc, (name, length, size) = params
            model = Sequential()
            model.add(InputLayer(input_shape=(length,), name=name))
            if type(size) == int:
                len_voc = max(voc.values()) + 1
                embedding = Embedding(len_voc, size)
            elif type(size) in [str, tuple] or type(size).__module__ == 'numpy':
                if type(size) != tuple: size = (size,)
                embedding = initemb(voc, *size)
            if size is not None: model.add(embedding)
            model.add(Flatten())
            return model

        # embeddings layer
        model = embs(embeddings, mkemb)

        # hidden layer
        if hidden is None:
            hidden = model.outputs[0]._keras_shape[1]
        model.add(Dense(hidden, activation=activation))
        model.add(Dropout(0.5))

        # softmax layer
        model.add(Dense(len(self.classes), bias=False, activation='softmax' if softmax else 'linear'))
        model.compile(loss=loss, optimizer=optimizer)
        return model
