from .joint import JointClassifier

class VoteClassifier(JointClassifier):
    def __init__(self, *parts):
        self.parts, self.weights = zip(*parts)

    def x_to_modelspace(self, x):
        return [part.x_to_modelspace(x) for part in self.parts]
    def y_to_modelspace(self, y):
        return [part.y_to_modelspace(y) for part in self.parts]

    def score(self, x, classes=None):
        each = [part.score(x, classes) for part, x in zip(self.parts, x)]
        raise NotImplementedError
        return self.parts[x[0]].score(x[1], classes[1])
    def score_on_batch(self, batch, classes=None):
        fs = [part.score_on_batch for part in self.parts]
        raise NotImplementedError
        return apply_on_batch_parts(fs, *batch)

    def predict(self, x, classes=None, argmax=True, scores=False, raw=False):
        from itertools import groupby
        each = [part.predict(x, classes, argmax, scores, raw) for part, x in zip(self.parts, x)]
        if scores: each = [x[0] for x in each]
        score = {c:sum(v[1] for v in v) for c,v in groupby(sorted(list(zip(each, self.weights))), key=lambda x:x[0])}
        #
        out = max(score.items(), key=lambda x:x[1])[0]
        from collections import defaultdict
        score2 = defaultdict(float)
        if scores:
            for c,v in score.items(): score2[self.parts[0].classes[c]] = v
            out = out, score2
        return out
    def update(self, y, _y, _x, x=(0,None), extra=None):
        self.parts[y[0]].update(y[1], _y[1], _x[1], x[1], extra)

    def predict_on_batch(self, batch, classes=None, argmax=True, scores=False, raw=False):
        raise Exception
        fs = [part.predict_on_batch for part in self.parts]
        #
        return apply_on_batch_parts(fs, *batch)
    def train_on_batch(self, *batch):
        raise Exception
        fs = [part.train_on_batch for part in self.parts]
        return apply_on_batch_parts(fs, *batch, indexed=4)

def apply_on_batch_parts(fs, *batch, indexed=1):
    from misc.itertools import deinterleave
    batch = [arg if type(arg) == list else [arg] * len(batch[0]) for arg in batch]
    batch = batch[:indexed] + [[(None,v) for v in arg] for arg in batch[indexed:]]
    interleaved = [(v[0][0], [arg[1] for arg in v]) for v in zip(*batch)]
    interleaved = [(i, (pos,v)) for pos,(i,v) in enumerate(interleaved)]
    deinterleaved = deinterleave(interleaved, len(fs))
    as_subbatch = [tuple(zip(*pos_subbatch)) for pos_subbatch in deinterleaved]
    as_subbatch = [(pos_subbatch[0], map(list,zip(*pos_subbatch[1]))) if pos_subbatch else ((),()) for pos_subbatch in as_subbatch]
    applied = [(v[0], f(*v[1]) if v[1] else []) for f,v in zip(fs,as_subbatch)]
    applied = [(positions, outputs if outputs else tuple(None for _ in positions)) for positions, outputs in applied]
    outputs = [x for positions, outputs in applied for x in zip(positions, outputs)]
    return [out for _,out in sorted(outputs)]
