import abc

from structured_prediction.AbstractStructuredPredictor import AbstractStructuredPredictor

SHIFT, RIGHT, LEFT, REDUCE = 0, 1, 2, 3
INVERT_MOVE = {SHIFT: "SHIFT", RIGHT: "RIGHT", LEFT: "LEFT", REDUCE: "REDUCE"}

# helper method for features' extraction
def pad_list(lst, N=3, default=None):
    return lst[:N] + [default] * (N - len(lst))

class Composable(abc.ABCMeta):
    def __add__(self, other):
        system = self.__module__.split('.')[-1]
        implem = self.__name__
        details = other.__name__.replace('Parser','').replace('Perceptron','').replace('FeedForward','')
        return type(system+details+implem, (self, other), {})

class AbstractParser(AbstractStructuredPredictor, metaclass=Composable):

    root_first = False
    joint = False

    def __init__(self, feature_template, moves):
        self.moves = moves
        self.features = feature_template

    def initialize(self, dataset, *args, **kwargs):
        self._initmodel(dataset, *args, **kwargs)
        return self

    def with_root_first(self, root_first=True):
        self.root_first = root_first
        return self

    @abc.abstractmethod
    def search(self, blackbox, sample, **kwargs): pass
    def xs(self, sample, **kwargs):
        def xs(blackbox):
            return self.search(blackbox, sample, **kwargs)
        return xs

    @abc.abstractmethod
    def learn(self, blackbox, sample, **kwargs): pass
    def xys(self, sample, **kwargs):
        def xys(blackbox):
            return self.learn(blackbox, sample, **kwargs)
        return xys


    constrained_predictions = True

    @staticmethod
    def polish(output):
        heads, _ = next(output)
        return heads

    @staticmethod
    def forget_gold(sample):
        heads, labels = (sample + (None, None))[2:4]
        heads = ((None,) + heads[1:]) if type(heads) == tuple else None
        labels = ((None,) + labels[1:]) if type(labels) == tuple else None
        return sample[:2] + (heads, labels)

    @staticmethod
    def _init_config(sample, root_first=False):
        words, tags, heads = sample[:3]
        extra = sample[3:]
        n = len(words) - 2

        if type(heads) != tuple:
            heads = heads, None
        gold, constraints = heads[:2]
        if not constraints:
            constraints = [None for _ in words]
        extra = ([None for _ in words] if not labels else labels for labels in extra)

        if root_first:
            from pan4.format import root_at_start
            sample = root_at_start(words, tags, gold, *extra)
            words, tags, gold = sample[:3]
            extra = sample[3:]
            _, _, constraints, _ = root_at_start(heads=constraints)

        def expand(heads):
            from pan4.preprocess import children, spans
            fuzzy = [+1 if type(h)==tuple and h[0]==-1 else -1 if type(h)==tuple and h[1]==-1 else None for h in heads]
            heads = [h if type(h) == int else None for h in heads]
            _children = children(heads)
            return heads, _children, spans(heads, _children), fuzzy

        if gold: gold = expand(gold)
        constraints = expand(constraints)

        return ((words, tags) + tuple(extra) + (root_first, n)), gold, constraints

    @staticmethod
    def _joint_lookahead(*a, **kw):
        return {}
        yield

    @staticmethod
    def _joint_learn(*a, **kw):
        return
        yield

    @staticmethod
    def _joint_wrap(span, *args): return tuple(args)
