from .AbstractBeamParse import AbstractBeamParse

import abc
import itertools

# 7-uples representing stack and buffer tokens
LEFTMOST_CHILD=0
SND_LEFTMOST_CHILD=1
LEFT_CHILDREN=2
ID=3
RIGHT_CHILDREN=4
SND_RIGHTMOST_CHILD=5
RIGHTMOST_CHILD=6

class BeamStackBufferState(AbstractBeamParse):

    def __init__(self,
                 _n0=(None,None,0,1,0,None,None),
                 _s0=(None,None,0,0,0,None,None),
                 _stack_tail=None,
                 _prev=(None,None,0)
                 ):
        self.n0 = _n0
        self.s0 = _s0
        self.stack_tail = _stack_tail
        super().__init__(_prev=_prev)

    def __str__(self):
        from .AbstractParser import INVERT_MOVE
        stack_tail = "\t\n" + "\n".join(["\t" + line for line in str(self.stack_tail).split("previous")[0].split('\n') if line.strip()[:2]=="s0"])
        previous = "\n".join(["\t" + line for line in str(self.previous).split('\n')])
        return "\n".join(["BeamParse ["] + ["\t" + line for line in [
                          "n0: {}".format(self.n0),
                          "s0: {}".format(self.s0),
                          "stack_tail: {}".format(stack_tail[1:]),
                          "last_move: {}".format(INVERT_MOVE[self.last_move] if self.last_move is not None else None),
                          "previous: {}".format(previous[1:])
                          ]] + ["]"])

    # derivation identity
    def __eq__(self, other):
        if type(other) is not type(self): return False
        return (    self.n0[ID]     == other.n0[ID]
                and self.last_move  == other.last_move
                and self.previous   == other.previous)

    # configuration identity
    def same(self, other):
        if type(other) is not type(self): return False
        if not (    self.n0[ID]     == other.n0[ID]
                and self.s0[ID]     == other.s0[ID]
                and self.s0h()      == other.s0h()):
            return False
        if self.stack_tail is not None:
            return self.stack_tail.same(other.stack_tail)
        else:
            return self.stack_tail == other.stack_tail

    def format_derivation(self):
        derivation = []
        for parse in self.predecessors:
            if parse.last_move is not None:
                derivation.append((parse.depth, parse.last_move, list(reversed(list(parse.stack()))), range(parse.n0[ID], self.sentence[-1] + 1)))
        from .AbstractParser import INVERT_MOVE
        return [""] + ["{}\td={}\t{}\t{}.   {}".format(i, d, INVERT_MOVE[move], " ".join(["["]+[str(s) for s in stack]+["]"]), " ".join(str(b) for b in buf)) for i, (d, move, stack, buf) in enumerate(reversed(derivation),start=1)] + ["===> {}".format(self.length_final)]

    def push(self):
        return self
    def pop(self):
        return self.stack_tail.stack_tail

    def has_buffer(self):
        return self.n0[ID] <= self.sentence[-1]

    # returns True if n0 is the last token of the buffer
    def end_buffer(self, n0=None):
        if n0 is None: n0 = self.n0[ID]
        return n0 == self.sentence[-1]

    # contains a concrete token
    def has_stack(self):
        return self.s0[ID]

    # each tail contains a concrete token
    @property
    def _stack_tails(self):
        sub_parse = self
        while sub_parse:
            if sub_parse.has_stack():
                yield sub_parse
            sub_parse = sub_parse.stack_tail

    @property
    def _stack(self):
        return (tail.s0 for tail in self._stack_tails)

    # concrete tokens only, cannot be 0
    def stack(self):
        return (s0[ID] for s0 in self._stack)

    # excludes 0 and its children
    def nohead_stack(self):
        return (tail.s0[ID] for tail in self._stack_tails if tail.s0h() is None)

    def s(self, i, idx=None):
        for n,s in enumerate(self._stack):
            if n==i: return s if idx is None else s[idx]
        return None

    # starting with s1; can be 0, not None
    @property
    @abc.abstractmethod
    def s0_ancestors(self):
        pass

    # either s1 (can be 0) or None
    def s0h(self):
        return next(self.s0_ancestors, None)

    def get_cut_edges(self, move, constraints=False):
        ref = self.constraints if constraints else self.gold
        root_first = self.sentence[-2]
        return self._get_cut_edges(move, self.n0[ID], self.s0[ID], self.stack, self.nohead_stack, ref, root_first, self.end_buffer, constraints=constraints) #

    @abc.abstractmethod
    def _get_cut_edges(self, *args, **kwargs):
        pass

    def extract_features(self):
        from .AbstractParser import pad_list

        words, tags = self.sentence[:2]

        def get_stack_context(data):
            s = self.stack()
            for _ in range(3):
                try:
                    yield data[next(s)]
                except:
                    yield None

        def get_parse_context(word, deps, data):
            if word == -1: deps = [0]
            valency = deps[-1]
            return itertools.islice(itertools.chain([valency], (data[d[0] if type(d) == tuple else d] for d in deps[:-1] if d), [None] * 2), 3)

        def get_label_context(deps):
            return (d[1] if type(d) == tuple else None for d in deps[:-1])

        def get_history():
            hist = self.predecessors
            for _ in range(8):
                try:
                    yield next(hist).last_move
                except:
                    yield None

        # The name of all features follows the following pattern:
        # - a letter: s --> stack n --> buffer
        # - an integer: the position on the stack or on the buffer
        # - an optionnal letter: l = left most child, r = rightmost child
        #   or h for the head
        # - a letter: p for POS tags and w for word identity

        s0 = self.s0[ID] if self.has_stack() else -1
        n0 = self.n0[ID]

        s0w, s1w, s2w = get_stack_context(words)
        s0p, s1p, s2p = get_stack_context(tags)

        n0w, n1w, n2w = pad_list(words[n0:])
        n0p, n1p, n2p = pad_list(tags[n0:])

        n0lv, n0lw, n0l2w = get_parse_context(n0, self.n0[:ID], words)
        n0lv, n0lp, n0l2p = get_parse_context(n0, self.n0[:ID], tags)
        n0lL, n0l2L = get_label_context(self.n0[:ID])

        n0rv, n0rw, n0r2w = get_parse_context(n0, self.n0[:ID:-1], words)
        n0rv, n0rp, n0r2p = get_parse_context(n0, self.n0[:ID:-1], tags)
        n0rL, n0r2L = get_label_context(self.n0[:ID:-1])

        s0lv, s0lw, s0l2w = get_parse_context(s0, self.s0[:ID], words)
        s0lv, s0lp, s0l2p = get_parse_context(s0, self.s0[:ID], tags)
        s0lL, s0l2L = get_label_context(self.s0[:ID])

        s0rv, s0rw, s0r2w = get_parse_context(s0, self.s0[:ID:-1], words)
        s0rv, s0rp, s0r2p = get_parse_context(s0, self.s0[:ID:-1], tags)
        s0rL, s0r2L = get_label_context(self.s0[:ID:-1])

        ancestors = self.s0_ancestors
        s0h, s0h2 = next(ancestors, None), next(ancestors, None)

        s0hw = words[s0h] if s0h else None
        s0hp = tags[s0h] if s0h else None

        s0h2w = words[s0h2] if s0h2 else None
        s0h2p = tags[s0h2] if s0h2 else None

        # distance between s0 and n0
        dist = n0 - s0 if s0 > 0 else 0 #

        hist_feat = get_history()

        return (s0w, s1w, s2w,
                s0p, s1p, s2p,
                n0w, n1w, n2w,
                n0p, n1p, n2p,
                n0lv, n0lw, n0l2w, n0lp, n0l2p, n0lL, n0l2L,
                n0rv, n0rw, n0r2w, n0rp, n0r2p, n0rL, n0r2L,
                s0lv, s0lw, s0l2w, s0lp, s0l2p, s0lL, s0l2L,
                s0rv, s0rw, s0r2w, s0rp, s0r2p, s0rL, s0r2L,
                s0hw, s0hp,
                s0h2w, s0h2p,
                dist, hist_feat)
