from .AbstractParser import AbstractParser

class CascadeParser(AbstractParser):
    def __init__(self, *a, parts=[], **kw):
        self.parts = list(parts)
        super().__init__(None, None, *a, **kw)

    def __add__(self, other):
        self.add(other)
        return self

    def __getitem__(self, *a, **kw): return self.parts.__getitem__(*a, **kw)
    def __delitem__(self, *a, **kw): return self.parts.__delitem__(*a, **kw)
    def add(self, part): return self.parts.append(part)
    def insert(self, *a, **kw): return self.parts.insert(*a, **kw)
    def pop(self, *a, **kw): return self.parts.pop(*a, **kw)

    def _defaultmodel(self, *a, **kw): pass
    def learn(self, blackbox, sample, **kwargs): pass
    def search(self, blackbox, sample, **kwargs): pass

    def process_sample(self, sample, batched=False, **kwargs):
        askbest = lambda hyp: iter([hyp])
        def constrain(ex):
            heads = ex[2] if type(ex[2])==tuple else (ex[2], [None for _ in ex[2]])
            return ex[:2] + ((heads[0], heads[-1]),) + ex[3:]
        def replace_out(ex, heads):
            #
            if type(heads) != tuple:
                heads = (ex[2][:2] if type(ex[2])==tuple else (ex[2],)) + (heads,)
            return ex[:2] + (heads,) + ex[3:]

        if batched:
            for part in self.parts:
                sample = [constrain(ex) for ex in sample]
                heads = part.process_sample(sample, batched=batched, **kwargs)
                sample = [replace_out(ex, next(h)[0]) for ex, h in zip(sample, heads)]
            return [askbest((ex[2][-1], None)) for ex in sample]
        for part in self.parts:
            sample = constrain(sample)
            heads = part.process_sample(sample, batched=batched, **kwargs)
            sample = replace_out(sample, next(heads)[0])
        return askbest((sample[2][-1], None))
