from .AbstractParser import AbstractParser

from .perceptron_parser import PerceptronParser

class JointParser(AbstractParser):

    joint = True

    def __init__(self, *a, main=PerceptronParser, **kw):
        self.main_class = main
        super().__init__(*a, **kw)

    def _defaultmodel(self, *a, **kw):
        from classifier.joint import JointClassifier
        main = self.main_class._defaultmodel(self, *a, **kw)
        return JointClassifier(*self._joint_models(main, *a, **kw))

    joint_names = []

    @staticmethod
    def _joint_models(main, dataset, *args, **kwargs): return (main,)

    def _joint_lookahead(self, blackbox, features): #
        extra = {}
        for i, name in enumerate(self.joint_names, start=1):
            yield True, joint(i, 1, features, None)
            _, (_, scores) = next(blackbox.next)
            # greedy decoding
            extra[name] = max(scores.items(), key=lambda x:x[1])[0]
        return extra

    def _joint_learn(self, prediction):
        for i, update in enumerate(self._joint_update_configs(), start=1):
            global_features = [update(parse, parse.sentence[i+1]) for parse in prediction.predecessors]
            for c in reversed(global_features):
                if c is None: # dummy for time averaging
                    y = _y = list(self.model.parts[i].classes)[0]
                    _X = (i, None)
                else:
                    y,_y,_X = c
                yield False, joint(i, 2, y, _y, _X, _X, {})

    @staticmethod
    def _joint_wrap(*args): return joint(0, *args)

def joint(part, span, *args):
    return tuple((part, arg) if i < span else arg for i, arg in enumerate(args))
