#
def maximum_spanning_tree(parses, projective=False, treshold=None):
    from collections import defaultdict
    candidates = defaultdict(float)
    for heads,weight in parses:
        for edge in enumerate(heads[1:-1],start=1):
            candidates[edge] += weight
    n = len(heads)
    comp = [i for i in range(n)]
    heads = [None for _ in comp]
    heads[-1] = 0
    def find(x):
        if x is None or comp[x] == x:
            return x
        comp[x] = find(comp[x])
        return comp[x]
    # check that edge k <-- i is compatible with other edges
    def check_projectivity(k,i):
        if i is None: return True
        for l,j in enumerate(heads[1:],start=1):
            if j is None:
                continue
            # non-projectivity condition for partial trees
            if (j == k and (j > i > l or j < i < l)) or (i == l and (i > j > k or i < j < k)):
                return False
            # canonical non-projectivity condition
            if min(i, k) < min(j, l) < max(i, k) < max(j, l):
                return False
        return True

    for c,h in sorted(candidates.keys(), key=lambda edge: candidates[edge], reverse=True):
        if h is None: continue
        if treshold is not None and candidates[c,h] < treshold:
            break
        if heads[c] is not None:
            continue
        comp_h = find(h)
        if c == comp_h or (projective and not check_projectivity(c,h)):
            continue
        heads[c] = h
        comp[c] = comp_h
    return heads

def reparse(testset, apply_model, parts, weights=None, test=False):
    if weights is None: weights = [1 for _ in parts]
    if type(weights) not in [list,tuple]: weights = [weights[p] for p in parts]
    annotated = (apply_model(p, testset) for p in parts)

    reweight = lambda data, w: ((ex[2], w) for ex in data)
    hypotheses = zip(*(reweight(*data) for data in zip(annotated, weights)))
    mst = map(maximum_spanning_tree, hypotheses)

    def replace_out(ex, heads):
        if test: heads = (ex[2][:2] if type(ex[2])==tuple else (ex[2],)) + (heads,)
        return ex[:2] + (heads,) + ex[3:]
    return (replace_out(ex, mst) for ex, mst in zip(testset, mst))
