from .AbstractParser import AbstractParser

import sys
sys.setrecursionlimit(10000)

print("""
[WARNING] The FeedForwardParser is only included as a code example: it is a WIP implementation that does not reach state-of-the-art accuracy.
[WARNING] We have not yet performed grid search, so you may reach good accuracy with appropriate hyperparameters, or it may contain actual issues.
[WARNING] The FeedForwardPOSTagger is fine.
[WARNING] Any feedback is welcome.""", file=sys.stderr)

def beam_loss(gold, pY):
    from keras import backend as K
    alpha, Y = gold[:,0], gold[:,1:]
    loss = - alpha * K.sum(pY*Y, axis=1)
    return K.sum(loss)

def cube(x):
    from keras import backend as K
    return x*x*x


class FeedForwardParser(AbstractParser):

    def _defaultmodel(self, dataset, *args, classes=None, embeddings=None, **kwargs):
        if classes is None: classes = self.moves
        from classifier.neural import weighted_linear_loss
        from classifier.feedforward import FeedForward
        pretrained = embeddings
        embeddings = [("word",16,10),("tag",16,10),("valencies",4,2),("distance",1,2),("history",8,2)]
        if pretrained: embeddings = [params[:2] + (pretrained.get(params[0], params[2]),) for params in embeddings]
        return FeedForward(classes,
            self._make_voc(dataset, tresh=4),
            embeddings=embeddings,
            hidden=40,
#            activation=cube, optimizer='adagrad',
            activation='tanh', optimizer='adagrad',
            softmax=False, loss=weighted_linear_loss)

    def _make_voc(self, sentences, tresh):
        START, END = "<start>", "ROOT"
        from collections import Counter
        words = [w for sent in sentences for w in sent[0]]
        words = [x[0] for x in filter(lambda x: x[1] >= tresh, Counter(words).most_common())]
        from conll.io import feat2tag
        tags = set(t for sent in feat2tag(sentences) for t in sent[1])
        return {
            "word": list(set([None, START, END] + list(words))),
            "tag": list(set([None, START, END] + list(tags))),
            "valencies": range(6),
            "distance": range(11),
            "history": [None] + list(self.moves),
            }
