from .AbstractParser import AbstractParser

class SubsampleParser(AbstractParser):
    def __init__(self, *a, gold=[], constraints=[], out=[], **kw):
        self.subsample_gold = gold if type(gold)==list else [gold]
        self.subsample_constraints = constraints if type(constraints)==list else [constraints]
        self.subsample_out = out if type(out)==list else [out]
        super().__init__(None, None, *a, **kw)

    def _defaultmodel(self, *a, **kw): pass
    def learn(self, blackbox, sample, **kwargs): pass
    def search(self, blackbox, sample, **kwargs): pass

    def process_sample(self, sample, batched=False, **kwargs):
        askbest = lambda hyp: iter([hyp])
        def replace_out(ex, heads):
            heads = (ex[2][:2] if type(ex[2])==tuple else (ex[2],)) + (heads,)
            return ex[:2] + (heads,) + ex[3:]
        def actions(ex, heads, i, action):
            heads_i = heads[i]
            if action[0] == "dropout":
                import random
                heads_i = [None if h and random.random() < action[1] else h for h in heads_i]
            if action[0] == "regions of bitags":
                tags = [t.cpos if 'cpos' in dir(t) else t for t in ex[1]]
                heads_i = [h if h is not None and '{}-{}'.format(t,tags[h]) in action[1] else None for c,(t,h) in enumerate(zip(tags, heads_i))]
            if action[0] == "regions of labels":
                labels = ex[3][0] if type(ex[3])==tuple else ex[3]
                heads_i = [h if l in action[1] else None for h,l in zip(heads_i, labels)]
            if action[0] == "maxlength":
                heads_i = [None if h and abs(c-h) > action[1] else h for c,h in enumerate(heads_i)]
            if action[0] == "minlength":
                heads_i = [None if h and abs(c-h) < action[1] else h for c,h in enumerate(heads_i)]
            if action[0] == "correct":
                heads_i = [h if gh==h else None for gh,h in zip(heads[0], heads_i)]
            if action[0] == "errors":
                heads_i = [None if h and gh==h else h for gh,h in zip(heads[0], heads_i)]
            return heads[:i] + (heads_i,) + (() if i==-1 else heads[i+1:])

        def subsample(ex):
            heads = ex[2]
            if type(heads) != tuple: heads = (heads,)
            for action in self.subsample_gold: heads = actions(ex, heads, 0, action)
            for action in self.subsample_constraints: heads = actions(ex, heads, 1, action)
            for action in self.subsample_out: heads = actions(ex, heads, -1, action)
            if type(ex[2]) != tuple: heads = heads[0]
            return heads

        if batched: return [askbest((subsample(ex), None)) for ex in sample]
        return askbest((subsample(sample), None))
