Pan4 format
===========

Represents a PoS- and dependency-annotated sentence, as a 4-uple `(words, tags, heads, labels)`.
All sequences are padded with 1 leading and 1 trailing tokens. Index 0 means the leading token and index 1 means the first word of the sentence. The `ROOT` token is always assumed to be placed at the end.

`words`:
    - a list of strings
        A tokenized sentence, preceded by a token '<start>' and followed by a token 'ROOT'.

`tags`: either...
    - a list of strings
        The part-of-speech tags for each token, preceded by '<start>' and followed by 'ROOT'.
    - a dict:
        XXX

`heads`: either...
    - None
        If no data to specify (typically when used as input).
    - a list of Nones and ints
        The dependency parse tree of the sentence.
        This bottom-up representation is a list of heads: the nth value is the head index of the nth token if it is known, and None otherwise. The first and last values (heads of '<start>' and 'ROOT') are always None and 0.
    - a (list of None and ints, list of lists of ints) tuple
        Both tuple members represent the same information in different forms, the first one is bottom-up, the second one is top-down.
        The bottom-up representation follows previous format.
        The top-down representation is a list of children lists: the nth value is the sorted list of the known children indexes of the nth token.
    - a 2-uple or 3-uple of any of the upper
        In that case the members can be either:
        - `(gold tree, head constraints)` (as input)
        - `(gold tree, predicted tree)` (as output)
        - `(gold tree, head constraints, predicted tree)`
        When constraints are specified, the first and last values (constraints on '<start>' and 'ROOT') must be None: the 'ROOT' token is always attached to '<start>' anyway.

`labels`: either...
    - None
        If no data to specify (typically when used as input).
    - a list of Nones and strs
        The dependency labels of the sentence.
        The nth value is the label of the dependency of the nth token to its head if it is known, and None otherwise.
        The first and last values are None.
    - a 2-uple or 3-uple of any of the upper
        In that case the members can be either
        - `(gold labels, label constraints)` (as input)
        - `(gold labels, predicted labels)` (as output)
        - `(gold labels, label constraints, predicted labels)`

Pan2 format
===========
Represents a PoS-annotated sentence, as a 2-uple `(words, tags)`.
Token sequences are not padded. Index 0 means the first word of the sentence.

`words`:
    - a list of strings
        A tokenized sentence, without padding.

`tags`: either...
    - a list of strings
        The part-of-speech tags for each token.
    - a dict:
        XXX

Morpho format
=============
XXX

BeamState format
================
Represents partial derivation hypotheses of the same state of the beam, as a k-best list.
The list is sorted by decreasing order of score.
It is guaranteed to be non-empty (no dead end hypotheses) and of size at most k.

Each hypothesis is an `AbstractBeamParse` instance, with two extra attributes:

`alive`: a boolean
    False iff the configuration is erroneous

`configuration`: a `(features, gold_moves, scores)` tuple
    - features: a defaultdict(int) with str keys
        The features extracted from this configuration.
    - gold_moves: a list of ints
         The zero-cost valid moves in this configuration.
    - scores: a list of floats
         For the valid move indexes, the scores computed at decoding time in this configuration; 0 otherwise.
