def zip_justify_compress(*iterables, fillvalue=None):
    from itertools import zip_longest
    z = zip_longest(*iterables, fillvalue=fillvalue)
    return (([x for x in zipped if x is not fillvalue], [x is not fillvalue for x in zipped]) for zipped in z)

def zip_justify(*iterables, fillvalue=None):
    from itertools import zip_longest
    return ([x for x in zipped if x is not fillvalue] for zipped in zip_longest(*iterables, fillvalue=fillvalue))

def multiplex(iterable, mkgen, batched=1):
    process = lambda batch: zip_justify(*list(map(mkgen,batch)))
    return batchmap(iterable, process, batched)

def batchmap(iterable, f, batched=1):
    from itertools import chain
    return chain.from_iterable(map(f, zip_justify(*[iterable]*batched)))

def uncompress(data, selectors, fillvalue=None):
    data = iter(data)
    return [next(data) if s else fillvalue for s in selectors]

#
def unzip_longest(iterable, fillvalue=None, batched=2):
    from itertools import tee, takewhile
    return [takewhile(lambda x: x is not fillvalue, map(lambda x,i=i:x[i],xs)) for i,xs in enumerate(tee(iterable,batched))]

def unzip_justify_compress(iterable, fillvalue=None, batched=2):
    return unzip_longest((uncompress(*x, fillvalue=fillvalue) for x in iterable), fillvalue=fillvalue, batched=batched)

def chunkpad(iterable, n=1, fillvalue=None):
    from itertools import islice
    while True:
        try:
            yield from islice(iterable, n)
            yield fillvalue
        except StopIteration:
            return

# debug utility
def gen_print(label, iterable, width=None):
    return ((print(label, str(x)[:width]),x)[1] for x in iterable)

def iassoc(iterable, key):
    return (v for k,v in iterable if k==key)

def deinterleave(iterable, keys=None):
    from itertools import tee
    if keys is None:
        keys = (True, False)
    elif type(keys) == int:
        keys = range(keys)
    return tuple(iassoc(*e) for e in zip(tee(iterable, len(keys)), keys))

def next0(iterable):
    for k,v in iterable:
        if not k: return v
        yield k,v

def catch0(iterable):
    forward, values = deinterleave(iterable)
    yield from ((True, val) for val in forward)
    return values
