class DefaultList(list):
    """
    A list that returns a default value if index out of bounds.
    """

    def __init__(self, what=None, default=None):
        what = what if what is not None else []
        self.default = default
        list.__init__(self, what)

    def __getitem__(self, index):
        try:
            return list.__getitem__(self, index)
        except IndexError:
            return self.default

import threading

class threadsafe_iter(object):
    """Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return self.it.__next__()

def threadsafe_generator(f):
    def g(*a, **kw):
        return threadsafe_iter(f(*a, **kw))
    return g

def multisampling(dataset):
    dataset = list(dataset)
    import random
    #
    def repeat(i,data):
        from itertools import tee
        data, dataset[i] = tee(data)
        from itertools import count
        for i in count():
            j = None
            for j,x in enumerate(data): yield i, x
            if j is None: break
    def stop_repeat(batches):
        for batch in batches:
            if all(x[0] for x in batch): break
            yield [x[1] for x in batch]
    for batch in stop_repeat(zip(*[repeat(*d) for d in enumerate(dataset)])):
        random.shuffle(batch)
        yield from batch

#########################
### INTERACTIVE DEBUG ###
#########################

def debug(sig, frame):
    import code, traceback
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d={'_frame':frame}         # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message  = "Signal received : entering python shell.\nTraceback:\n"
    message += ''.join(traceback.format_stack(frame))
    i.interact(message)

def listen():
    import signal
    signal.signal(signal.SIGUSR1, debug)
