def scores(test, label='', micro=True, minocc=0, minfreq=0):
    all_occ = sum(occ for _,occ in test.values())
    all_uas = sum(uas*occ for uas,occ in test.values())/all_occ
    print("")
    for cat,(uas,occ) in sorted(test.items()):
        if occ >= max(minocc, all_occ * minfreq / 100):
            print("[{}*]\t{}\t{:.2%} ({})".format(label,cat,uas,occ))
    if micro:
        print("{:.2%}".format(all_uas))

def uas_bitable(uas_table, minocc=10, minfreq=1, minocc_cell=10, minfreq_cell=10, occ=False, freq=False, maxlines=None, percent=True):
    all_rows = set(r for (r,_),_ in uas_table.items())
    all_cols = set(c for (_,c),_ in uas_table.items())

    all_occ = sum(occ for _,occ in uas_table.values())
    all_uas = sum(uas*occ for uas,occ in uas_table.values())/all_occ
    row_occ = {row: sum(occ for (r,_),(_,occ) in uas_table.items() if r==row) for row in all_rows}
    row_uas = {row: sum(uas*occ for (r,_),(uas,occ) in uas_table.items() if r==row)/row_occ[row] for row in all_rows}
    col_occ = {col: sum(occ for (_,c),(_,occ) in uas_table.items() if c==col) for col in all_cols}
    col_uas = {col: sum(uas*occ for (_,c),(uas,occ) in uas_table.items() if c==col)/col_occ[col] for col in all_cols}

    key = lambda x: x.replace("ROOT"," ROOT") if type(x)==str else x
    rows = [row for row in all_rows if row_occ[row] >= max(minocc, all_occ * minfreq / 100)]
    if maxlines: rows = sorted(rows, key=lambda row: row_occ[row], reverse=True)[:maxlines]
    rows.sort(key=key)
    cols = [col for col in all_cols if col_occ[col] >= max(minocc, all_occ * minfreq / 100)]
    cols.sort(key=key)


    l = max(len(str(r)) for r in rows)

    lines = []
    lines.append([""])

    lines.append(["".ljust(l)] + ["{}".format(c) for c in cols] + ["=TOT="])

    percent = '1%' if percent else '2f'
    for r in rows:
        row = list(map(lambda c: (len(str(c)), uas_table.get((r,c),(None,0))), cols))
        line = ["{}".format(r).ljust(l)]
        line += [("{:.%s}" % percent).format(uas).ljust(lc) if uas is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, (uas, occ) in row]
        line += [("{:.%s} ({})" % percent).format(row_uas[r], row_occ[r])]
        line += ["[{:.0%}]".format(row_occ[r]/all_occ)]
        lines.append(line)
        if occ:
            line = [".".ljust(l)]
            line += ["({})".format(occ).ljust(lc) if uas is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, (uas, occ) in row]
            lines.append(line)
        if freq:
            line = [".".ljust(l)]
            line += ["[{:.0%}]".format(occ/row_occ[r]).ljust(lc) if uas is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, (uas, occ) in row]
            lines.append(line)

    line1, line2, line3 = ["=TOT=".ljust(l)], ["".ljust(l)], ["".ljust(l)]
    for c in cols:
        line1 += [("{:.%s}" % percent).format(col_uas[c]).ljust(len(str(c)))]
        line2 += ["({})".format(col_occ[c]).ljust(len(str(c)))]
        line3 += ["[{:.0%}]".format(col_occ[c]/all_occ).ljust(len(str(c)))]

    line1 += [("{:.%s}" % percent).format(all_uas)]
    line2 += ["      ({})".format(all_occ)]

    for line in lines + [line1, line2, line3]:
        print('\t'.join(line))

def confusion_matrix(matrix, minocc=10, minfreq=1, minocc_cell=10, minfreq_cell=10, occ=True, freq=False, maxlines=None):
    all_rows = set(r for (r,_),_ in matrix.items())
    all_cols = set(c for (_,c),_ in matrix.items())

    all_occ = sum(occ for occ in matrix.values())
    row_occ = {row: sum(occ for (r,_),occ in matrix.items() if r==row) for row in all_rows}
    col_occ = {col: sum(occ for (_,c),occ in matrix.items() if c==col) for col in all_cols}

    key = lambda x: x.replace("ROOT"," ROOT") if type(x)==str else x
    rows = [row for row in all_rows if row_occ[row] >= max(minocc, all_occ * minfreq / 100)]
    if maxlines: rows = sorted(rows, key=lambda row: row_occ[row], reverse=True)[:maxlines]
    rows.sort(key=key)
    cols = [col for col in all_cols if col_occ[col] >= max(minocc, all_occ * minfreq / 100)]
    cols.sort(key=key)


    l = max(len(str(r)) for r in rows)

    lines = []
    lines.append([""])

    lines.append(["".ljust(l)] + ["{}".format(c) for c in cols] + ["=TOT="])

    for r in rows:
        row = list(map(lambda c: (len(str(c)), matrix.get((r,c))), cols))
        line = ["{}".format(r).ljust(l)]
        if occ or not freq:
            line += ["{}".format(occ).ljust(lc) if occ is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, occ in row]
        else:
            line += ["{:.0%}".format(occ/row_occ[r]).ljust(lc) if occ is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, occ in row]
        line += ["{}".format(row_occ[r])]
        line += ["[{:.0%}]".format(row_occ[r]/all_occ)]
        lines.append(line)
        if freq and occ:
            line = [".".ljust(l)]
            line += ["[{:.0%}]".format(occ/row_occ[r]).ljust(lc) if occ is not None and occ >= max(minocc_cell, row_occ[r] * minfreq_cell / 100) else "" for lc, occ in row]
            lines.append(line)

    line1, line2 = ["=TOT=".ljust(l)], ["".ljust(l)]
    for c in cols:
        col = map(lambda r: matrix.get((r,c)), rows)
        line1 += ["{}".format(col_occ[c])]
        line2 += ["[{:.0%}]".format(col_occ[c]/all_occ)]

    line1 += ["{}".format(all_occ)]

    for line in lines + [line1, line2]:
        print('\t'.join(line))
