Training strategy
=================

- oracle: an int, or None
    If not None, the combined oracle/training base strategy identifier (static or dynamic, early update or max-violation...).

- exploration_rate: a float (default: 1)
    After an update, search is resumed on the prediction or the reference with probability `exploration_rate` and `1 - exploration_rate`.
    A value under 1 enables training for error recovery.
    Used by the `_infer` function.

- beam_update: a boolean (default: False)
    Experimental feature.
    If True, in train mode, do not update non-final configurations against the 1-best hypothesis but against the last hypothesis in the beam.

- check_cut_edge: a boolean (default: True)
    Controls dynamic oracles.
    Flag (parse-level) errors when an action cuts a gold edge that was still reachable.
    Used by the `lookahead` function.

- check_single_reference: (A) a boolean (default: False); or (B) an instance of AbstractBeamParse, or None (default: None)
    Controls static oracles.
    Flag (parse-level) errors when an action differs from (A) a precomputed derivation (B) `check_single_reference`'s action.

- time_limit: an int, or None (default: None)
    If not None, allow only `time_limit` extension steps of the beam.
    Used by the `lookahead` function.

- remember_bad_beams: a float (default: 1)
    For the error found, remember the state of the beam when the first `remember_bad_beams` (in %) part of the beam was in error.
    It is reinitialized if an error-to-be is corrected.
    Used by the `lookahead` function.

- restart: a boolean (default: True)
    Triggers a restart strategy after each update: decoding is restarted in the positive or the negative example, until next error.

- batch_sentence: a boolean (default: False)
    For given sentence, do all necessary decodings before performing updates.

- no_final_ranking: a boolean
    Experimental feature.
    In train mode, do not yield states still containing a reference in the beam; interrupt decoding instead.
    This should prevent model updates when the error is a bad ranking of final configurations, thus reducing the beam purpose to keeping references inside the beam.

Subsampling
===========

- p_no: a 4-uple of floats (default: (0,0,0,0))
    API for experimenting with probability tresholds.
    When an update has been marked as necessary, it is ignored with probability `p_no[2]` and training goes on as if it had been done.

- p_reset: a float
    Probability to interrupt current decoding and restart as if (potentially erroneous) current 1-best was the initial configuration.
    Not done on final configurations nor when a state-level error has already been reached.
    Non-null values enforce training in suboptimal configurations sampled from the decoding space.
    Used by the `lookahead` function.

Monitoring
==========

- compute_coverage: a boolean (default: False)
    Display after each epoch the accuracy on decisions taken on first decoding of each training sentence.
    This option implies a computation time overhead.

- feature_file: a filehandler, or None (default: None)
    If not None, dump features for each evaluated configuration in `feature_file`.
