def project_parse(src, tgt, align):
    from collections import defaultdict
    _, src_tags, src_heads, src_labels = src[:4]
    _, tgt_tags, _, _ = tgt[:4]
    if not src_labels: src_labels = [None for _ in src_heads]
    tgt_heads = [None for _ in tgt_tags][:-1] + [0]
    tgt_labels = [None for _ in tgt_tags]
    src_tags = [t.cpos if 'cpos' in dir(t) else t for t in src_tags]
    tgt_tags = [t.cpos if 'cpos' in dir(t) else t for t in tgt_tags]
    als = defaultdict(list)
    alt = defaultdict(list)
    for (s,t) in align:
        if src_tags[s] != tgt_tags[t]: continue
        als[s].append(t)
        alt[t].append(s)
    for s_child,s in enumerate(src_heads):
        if not len(als[s_child]) == len(als[s]) == 1: continue
        t_child, t = als[s_child][0], als[s][0]
        if not len(alt[t_child]) == len(alt[t]) == 1: continue
        tgt_heads[t_child] = t
        tgt_labels[t_child] = src_labels[s_child]

    return tgt[:2] + (tgt_heads, tgt_labels)

def reorder_sentence(src, tgt, align):
    from collections import defaultdict
    src_words, src_tags = src[:2]
    tgt_words, tgt_tags = tgt[:2]
    tgt_words = ["<start>"] + ["_" for _ in tgt_tags][1:-1] + ["ROOT"]
    als = defaultdict(list)
    alt = defaultdict(list)
    for (s,t) in align:
        if src_tags[s] != tgt_tags[t]: continue
        als[s].append(t)
        alt[t].append(s)
    for s,w in enumerate(src_words):
        if not len(als[s]) == 1: continue
        t = als[s][0]
        if not len(alt[t]) == 1: continue
        tgt_words[t] = w

    return (tgt_words,) + tgt[1:]

def align_parse(src, tgt, align, coeff_triangle=0.25):
    from collections import defaultdict
    src_heads = src[2]
    tgt_heads = tgt[2]
    als = defaultdict(list)
    alt = defaultdict(list)
    for (s,t) in align:
        als[s].append(t)
        alt[t].append(s)
    score = 0
    for s_child,s in enumerate(src_heads):
        n_squares = sum(1 for t_child in als[s_child] if tgt_heads[t_child] in als[s])# or t_child in als[s] or tgt_heads[t_child] in als[s_child])
        n_triangles = sum(1 for t_child in als[s_child] if tgt_heads[t_child] not in als[s] and t_child in als[s])
        n_mismatches = len(als[s_child])-(n_squares+n_triangles)
        score += n_squares/2 + coeff_triangle * n_triangles - n_mismatches
    for t_child,t in enumerate(tgt_heads):
        n_squares = sum(1 for s_child in alt[t_child] if src_heads[s_child] in alt[t])# or s_child in alt[t] or src_heads[s_child] in alt[t_child])
        n_triangles = sum(1 for s_child in alt[t_child] if src_heads[s_child] not in alt[t] and s_child in alt[t])
        n_mismatches = len(alt[t_child])-(n_squares+n_triangles)
        score += n_squares/2 + coeff_triangle * n_triangles - n_mismatches
    return score
