#!/usr/bin/python3

CPOS, WORD, CHAR = range(3)

def ngram_dist(dataset, size=3, minsize=None, data=CPOS):
    from collections import defaultdict
    from pan4.format import pad_tokens
    if minsize is None: minsize = size
    ttt_counts = defaultdict(int)
    for words, features, _, _ in dataset:
        if data == CPOS:
            tokens = [feats.cpos for feats in features]
            for _ in range(2, size): pad_tokens(tokens)
        elif data == WORD:
            tokens = list(words)
            for _ in range(2, size): pad_tokens(tokens)
        elif data == CHAR:
            tokens = " ".join(["<" * (size-2)] + words[1:-1] + [">" * (size-2)])
        for s in reversed(range(minsize, size+1)):
            for ttt in zip(*(tokens[i:] for i in range(s))):
                ttt_counts[ttt] += 1
            tokens = tokens[1:-1]
    return ttt_counts

def klcpos3(load_data, sources, target=None, size=3, minsize=None, data=CPOS, infile=None, outfile=None):
    if infile:
        allkl = dict()
        for line in infile:
            line = line.split("\t")
            src, tgt, div = line[0], line[1][4:], float(line[2][2:])
            allkl.setdefault(tgt,{})[src] = div
        return allkl
    if not outfile: from sys import stdout as outfile

    if target is not None: load_data = partial(load_data, target)
    distribution = { src: ngram_dist(load_data(src), size, minsize, data) for src in sources }
    ngram_index = dict()
    for _ in (ngram_index.setdefault(ttt, len(ngram_index)) for src in sources for ttt in distribution[src]): pass

    import numpy as np
    arrays = dict()
    for src in sources:
        arr = np.ones(len(ngram_index)) ###
        for ttt,occ in distribution[src].items():
            arr[ngram_index[ttt]] = occ + 1 ###
        arrays[src] = arr / np.sum(arr)

    allkl = dict()
    for tgt in (sorted(sources) if target is None else [target]):
        p = arrays[tgt]
        kl = { src: np.sum(p * np.log10(p/q)) for src in sources for q in (arrays[src],) }
        for src,score in sorted(kl.items(), key=lambda x:x[1]):
            print("{}\t--> {}\t| {:.3f}".format(src, tgt, score), file=outfile)
        allkl[tgt] = kl
    if target is not None: return allkl[target]
    return allkl
