from .postagger import AbstractPOSTagger, START, END, spelling_pattern

import sys
sys.setrecursionlimit(10000)

class FeedForwardPOSTagger(AbstractPOSTagger):

    def _defaultmodel(self, dataset):
        from classifier.neural import weighted_linear_loss
        from classifier.feedforward import FeedForward
        return FeedForward(self.classes,
            self._make_voc(dataset, tresh=5),
            embeddings=[("word",5,50),("tag",2,10),("sp",1,5),("pref",25,10),("suffix",35,10)],
            hidden=200,
            activation='relu', optimizer='adagrad')

    def _make_voc(self, sentences, tresh):
        from collections import Counter
        words = [self._normalize(w) for sent in sentences for w in sent[0]]
        words = [x[0] for x in filter(lambda x: x[1] >= tresh, Counter(words).most_common())]
        tags = set(t for sent in sentences for t in sent[1])
        sp = set(spelling_pattern(w) for sent in sentences for w in sent[0])
        prefixes = [w[:j] for w in START + END + words for j in range(1,6)]
        suffixes = [w[-j:] for w in START + END + words for j in range(1,8)]
        return {
            "word": list(set(START + END + list(words))),
            "tag": list(set(START + END + list(tags))),
            "sp": list(sp),
            "pref": list(set(prefixes)),
            "suffix": list(set(suffixes)),
            }

    def _get_features(self, i, word, context, prev, prev2):

        return {
            "word": context[i-2:i+3],
            "tag": [prev, prev2],
            "sp": [spelling_pattern(word)],
            "pref": [w[:j] for w in context[i-2:i+3] for j in range(1,6)],
            "suffix": [w[-j:] for w in context[i-2:i+3] for j in range(1,8)],
            }
