from .postagger import AbstractPOSTagger, spelling_pattern

class PerceptronPOSTagger(AbstractPOSTagger):

    def _defaultmodel(self, dataset):
        from classifier.avg_perceptron import Perceptron
        return Perceptron(self.classes)

    def _get_features(self, i, word, context, prev, prev2):
        """Map tokens into a feature representation, implemented as a
        {hashable: float} dict."""
        def add(name, *args, lex=0):
            try:
                if self.lexicalization_level < lex: return
            except AttributeError:
                pass
            try:
                import random
                if self.lexicalization_dropout and random.random() < 0.1 + 0.2*lex: return
            except AttributeError:
                pass
            try:
                if self.authorized_feats and set(name.split()) & set('word suffix pref'.split()) and ' '.join((name,) + tuple(args)) not in self.authorized_feats: return
            except AttributeError:
                pass
            features[' '.join((name,) + tuple(args))] = 1

        from collections import defaultdict
        features = defaultdict(int)

        add('bias')
        add('i sp', spelling_pattern(word), lex=1)
        for k,v in [('i',i),('i-1',i-1),('i-2',i-2),('i+1',i+1),('i+2',i+2)]:
            add(k + ' word', context[v], lex=4 if v==i else 3)
            for j in range(1,8):
                add(k + ' suffix', str(j), context[v][-j:], lex=2)
            for j in range(1,6):
                add(k + ' pref', str(j), context[v][:j], lex=2)
        add('i-1 tag', prev)
        add('i-2 tag', prev2)
        add('i-1 tag+i-2 tag', prev, prev2)
        add('i-1 tag+i word', prev, context[i], lex=4)

        return features
