import abc

class AbstractTransitionSystem(metaclass=abc.ABCMeta):
    def valid_moves(self, *args):
        return self._valid_moves(*self._valid_moves_elements(*args))
    @abc.abstractmethod
    def _valid_moves_elements(self, *args): pass

    @abc.abstractmethod
    def _valid_moves(self, *args): pass

    @abc.abstractmethod
    def _get_cut_edges(self, *args, **kwargs): pass

    @staticmethod
    def _auxcost_action(n0, _stack, _nohead_stack, head, children):
        def head_in_stack(c, start=0, end=None):
            if c is None or head[c] is None: return
            for i, s in enumerate(_stack()):
                if end is not None and i > end: break
                if s < head[c]: break
                if i < start: continue
                if s == head[c]: yield 1

        def head_in_buffer(c, start=0):
            if c is None or head[c] is None: return
            if head[c] >= n0 + start:
                yield 1

        def children_in_stack(h, start=0):
            for i, s in enumerate(_nohead_stack()):
                if i < start: continue
                if head[s] == h: yield 1

        def children_in_buffer(h, start=0):
            if h is None: return
            for child in children[h]:
                if child >= n0 + start: yield 1

        return head_in_stack, head_in_buffer, children_in_stack, children_in_buffer
