from .ArcEager import TransitionSystem as _TransitionSystem, BeamParse as _BeamParse, BeamParser as _BeamParser, LocalParser as _LocalParser

# unchecked
from .ArcEager import nomultiroot

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE

class TransitionSystem(_TransitionSystem):

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_s0h, has_buffer, end_buffer, has_stack, nohead_stack):
        if has_buffer():
            yield SHIFT
            if root_first or has_stack():
                yield RIGHT
        if has_stack():
            yield REDUCE
            if (not root_first or has_buffer()) and not has_s0h():
                yield LEFT

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, constraints=False):
        _, head_in_buffer, _, _ = self._auxcost_action(root_first, n0, _stack, _nohead_stack, *ref[:2])

        head = ref[0]
        def compulsory_none(c):
            if head[c] is None and not constraints: yield 1
        def s0_not_attached():
            return next(_nohead_stack(), None) == s0 #

        if move == SHIFT:
            pass
        if move == LEFT:
            yield from compulsory_none(s0)
        if move == RIGHT:
            yield from compulsory_none(n0)
        if move == REDUCE:
            if s0_not_attached():
                yield from head_in_buffer(s0)

        yield from super()._get_cut_edges(move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, constraints=False)

################################################################################
##                                AUXILIARY COST                              ##
################################################################################

    @staticmethod
    def _auxcost_arcdecomposability(n0, _stack, _nohead_stack, ref, root_first, _end_buffer):
        return lambda *a,**kw: iter([]), lambda *a,**kw: iter([])

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, _BeamParse):
    pass

class BeamParser(_BeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

class LocalParser(TransitionSystem, _LocalParser):
    pass
