from .ArcSpan import BeamParser as MixinBeamParser, LocalParser as MixinLocalParser
from .ArcEagerPartial import TransitionSystem as _TransitionSystem, BeamParse as _BeamParse, BeamParser as _BeamParser, LocalParser as _LocalParser
from .ArcEager import TransitionSystem as __TransitionSystem

# unchecked
from .ArcEagerPartial import nomultiroot

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE

class TransitionSystem(_TransitionSystem):

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(span, easy_span, root_first, has_s0h, has_buffer, end_buffer, has_stack, nohead_stack):
        from itertools import chain
        arcdecomposable = easy_span or not (root_first or nomultiroot)
        has_extended_stack = lambda: root_first or has_stack()
        x = False
        if has_buffer():
            # a SHIFT calls for a LEFT
            if arcdecomposable or not (end_buffer() and has_extended_stack()):
                x = True
                yield SHIFT
            if root_first or has_stack():
                stack_components = chain(nohead_stack(), [0] if root_first else [])
                # do not use last LEFT candidate
                if arcdecomposable or not (end_buffer() and len_gt(stack_components, 1)):
                    if not easy_span or span <= easy_span:
                        x = True
                        yield RIGHT
        if has_stack():
            if has_s0h() or easy_span:
                yield REDUCE
                x = True
            if not has_s0h() and (not root_first or has_buffer()):
                if not easy_span or span <= easy_span:
                    yield LEFT
                    x = True
        if not x:
            raise Exception(str((span, easy_span, root_first, has_s0h(), has_buffer(), end_buffer(), has_stack())))

################################################################################
##                                AUXILIARY COST                              ##
################################################################################

    @staticmethod
    def _auxcost_arcdecomposability(*args, **kwargs):
        import transition_system.ArcEager
        easy_span = 2 #
        return (_TransitionSystem if easy_span else __TransitionSystem)._auxcost_arcdecomposability(*args, **kwargs)

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

#
def BeamParseFactory(easy_span=0):
    class BeamParse(TransitionSystem, _BeamParse):
        def _valid_moves_elements(self):
            from dparser.BeamStackBufferState import ID
            span = abs(self.s0[ID] - self.n0[ID])
            return (span, easy_span) + super()._valid_moves_elements()
    return BeamParse

class BeamParser(MixinBeamParser, _BeamParser):
    def _defaultBeamParseFactory(self):
        return BeamParseFactory

class LocalParser(TransitionSystem, MixinLocalParser, _LocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        s0 = stack[-1] if stack else 0
        span = abs(s0 - n0)
        return (span, self.easy_span) + super()._valid_moves_elements(n0, n, stack, heads)
