from dparser.AbstractParser import AbstractParser

from .ArcHybrid import TransitionSystem as _TransitionSystem, BeamParse as _BeamParse, BeamParser as _BeamParser, LocalParser as _LocalParser

# unchecked
from .ArcHybrid import nomultiroot

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE
MOVES = (SHIFT, RIGHT, LEFT, REDUCE)

class Parser(AbstractParser):
    def __init__(self, features, *a, **kw):
        AbstractParser.__init__(self, features, MOVES, *a, **kw)

class TransitionSystem(_TransitionSystem):

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_token_s1, has_buffer, has_stack):
        if has_buffer():
            yield SHIFT
        if has_stack():
            if root_first or has_token_s1():
                yield RIGHT
            if not (root_first or nomultiroot and has_token_s1()) or has_buffer():
                yield LEFT
            yield REDUCE

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, constraints=False):
        head_in_extended_stack, head_in_buffer, _, children_in_buffer = self._auxcost_action(root_first, n0, _stack, _nohead_stack, *ref[:2])

        head = ref[0]
        def compulsory_none(c):
            if head[c] is None and not constraints: yield 1

        if move == SHIFT:
            pass
        if move == LEFT:
            yield from compulsory_none(s0)
        if move == RIGHT:
            yield from compulsory_none(s0)
        if move == REDUCE:
            yield from head_in_buffer(s0)
            yield from head_in_extended_stack(s0, 1, 1)
            yield from children_in_buffer(s0)

        yield from super()._get_cut_edges(move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, constraints=False)

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, _BeamParse):
    def transition(self, move, score=0, label=None):
        if move == REDUCE:
            n0 = self.n0
            s0 = self.stack_tail.s0
            tail = self.pop()
            new = self.__class__(n0, s0, tail, (self,move,score))
            if label is not None: new.last_label = label
            return new
        return super().transition(move, score=score, label=label)

class BeamParser(Parser, _BeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

class LocalParser(TransitionSystem, Parser, _LocalParser):
    def transition(self, move, n0, stack, parse):
        if move == REDUCE:
            stack.pop()
            return n0
        return super().transition(move, n0, stack, parse)
