from .ArcSpan import BeamParser as MixinBeamParser, LocalParser as MixinLocalParser
from .ArcHybridPartial import TransitionSystem as _TransitionSystem, BeamParse as _BeamParse, BeamParser as _BeamParser, LocalParser as _LocalParser

# unchecked
from .ArcHybridPartial import nomultiroot

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE

class TransitionSystem(_TransitionSystem):

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(spanL, spanR, easy_span, root_first, has_token_s1, has_buffer, has_stack):
        if has_buffer():
            yield SHIFT
        if has_stack():
            if root_first or has_token_s1():
                if not easy_span or spanL <= easy_span:
                    yield RIGHT
            if not (root_first or nomultiroot and has_token_s1()) or has_buffer():
                if not easy_span or spanR <= easy_span:
                    yield LEFT
            yield REDUCE

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

def BeamParseFactory(easy_span=0):
    class BeamParse(TransitionSystem, _BeamParse):
        def _valid_moves_elements(self):
            from dparser.BeamStackBufferState import ID
            s1 = self.s(1,ID)
            if s1 is None: s1 = 0
            spanL = abs(s1 - self.s0[ID])
            spanR = abs(self.s0[ID] - self.n0[ID])
            return (spanL, spanR, easy_span) + super()._valid_moves_elements()
    return BeamParse

class BeamParser(MixinBeamParser, _BeamParser):
    def _defaultBeamParseFactory(self):
        return BeamParseFactory

class LocalParser(TransitionSystem, MixinLocalParser, _LocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        s1 = stack[-2] if len(stack) >= 2 else 0
        s0 = stack[-1] if stack else 0
        spanL = abs(s1 - s0)
        spanR = abs(s0 - n0)
        return (spanL, spanR, self.easy_span) + super()._valid_moves_elements(n0, n, stack, heads)
