from dparser.AbstractParser import AbstractParser
from dparser.AbstractBeamParser import AbstractBeamParser
from dparser.AbstractLocalParser import AbstractLocalParser

import abc

class Parser(AbstractParser):
    easy_span = 2

    def with_easy_span(self, easy_span):
        self.easy_span = easy_span
        return self

    def forget_over_span(self, heads):
        if not self.easy_span or not heads: return heads
        return [None if h and abs(c-h) > self.easy_span else h for c,h in enumerate(heads)]

    def learn(self, blackbox, sample, **kwargs):
        # just in case
        if self.easy_span:
            heads = sample[2]
            gh = heads if type(heads) == tuple else (heads, None)
            gh = list(map(self.forget_over_span, gh[:2]))
            heads = gh + heads[2:] if type(heads) == tuple else gh[0]
            sample = sample[:2] + (heads,) + sample[3:]
        return super().learn(blackbox, sample, **kwargs)

    def search(self, *args, **kwargs):
        from misc.itertools import catch0
        kbest = yield from catch0(super().search(*args, **kwargs))
        # just in case
        if self.easy_span:
            kbest = ((self.forget_over_span(heads), hist) for heads, hist in kbest)
        yield from ((False, p) for p in kbest)

class BeamParser(Parser, AbstractBeamParser):

    def _defaultBeamParse(self):
        return self._defaultBeamParseFactory()(self.easy_span)()

    @abc.abstractmethod
    def _defaultBeamParseFactory(self):
        pass

class LocalParser(Parser, AbstractLocalParser):
    pass
