from .ArcEager import TransitionSystem as _TransitionSystem, BeamParse as _BeamParse, BeamParser as _BeamParser, LocalParser as _LocalParser

from .ArcEager import nomultiroot

from dparser.AbstractParser import SHIFT, RIGHT, LEFT, REDUCE

class TransitionSystem(_TransitionSystem):

################################################################################
##                                 LEGAL MOVES                                ##
################################################################################

    @staticmethod
    def _valid_moves(root_first, has_buffer, has_stack, has_stacks):
        if has_buffer():
            yield SHIFT
            if root_first or has_stack():
                yield RIGHT
        if has_stack():
            if root_first or has_stacks():
                yield REDUCE
            if not root_first or has_buffer():
                yield LEFT

################################################################################
##                                CORRECT MOVES                               ##
################################################################################

    def _get_cut_edges(self, move, n0, s0, _stack, _nohead_stack, ref, root_first, _end_buffer, **kwargs):
        head_in_extended_stack, head_in_buffer, children_in_stack, children_in_buffer = self._auxcost_action(root_first, n0, _stack, _stack, *ref[:2])
        _, s0_leftmost, _, s0_rightmost, left_headed, right_headed = self._auxcost_fuzzy(_stack, root_first, _end_buffer, ref[3])

        if move == SHIFT:
            yield from head_in_extended_stack(n0, 1)
            yield from children_in_stack(n0)
        if move == LEFT:
            yield from head_in_buffer(s0, 1)
            yield from head_in_extended_stack(s0, 1, 1)
            yield from children_in_buffer(s0)
            if s0_leftmost():
                if left_headed(n0): yield 1
            else:
                if left_headed(s0): yield 1
        if move == RIGHT:
            yield from head_in_extended_stack(n0, 1)
            yield from children_in_stack(n0)
        if move == REDUCE:
            yield from head_in_buffer(s0)
            yield from children_in_buffer(s0)
            if right_headed(s0) and not s0_rightmost(): yield 1
            if left_headed(n0) and s0_leftmost(): yield 1

################################################################################
##                                MOVE SEMANTICS                              ##
################################################################################

class BeamParse(TransitionSystem, _BeamParse):
    def _valid_moves_elements(self):
        root_first = self.sentence[-2]
        has_buffer = self.has_buffer
        has_stack = self.has_stack
        has_stacks = lambda: self.s(1) is not None
        return root_first, has_buffer, has_stack, has_stacks

    def transition(self, move, score=0, label=None):
        from dparser.BeamStackBufferState import LEFTMOST_CHILD, SND_LEFTMOST_CHILD, LEFT_CHILDREN, ID, RIGHT_CHILDREN, SND_RIGHTMOST_CHILD, RIGHTMOST_CHILD
        monotonic = super().transition(move, score, label)
        if move == LEFT:
            # amend RIGHT
            if monotonic.s0[RIGHTMOST_CHILD] == self.s0[ID]:
                #
                new = self.__class__(
                    monotonic.n0,
                    monotonic.s0[:ID+1] + (monotonic.s0[RIGHT_CHILDREN]-1, None, monotonic.s0[SND_RIGHTMOST_CHILD]),
                    monotonic.stack_tail,
                    (self,move,score))
                if 'last_label' in monotonic.__dict__: new.last_label = monotonic.last_label
                return new
        elif move == REDUCE:
            # amend SHIFT
            if monotonic.s0[RIGHTMOST_CHILD] != self.s0[ID]:
                new = self.__class__(
                    monotonic.n0,
                    monotonic.s0[:ID+1] + (monotonic.s0[RIGHT_CHILDREN]+1, monotonic.s0[RIGHTMOST_CHILD], self.s0[ID]),
                    monotonic.stack_tail,
                    (self,move,score))
                if 'last_label' in monotonic.__dict__: new.last_label = monotonic.last_label
                return new
        return monotonic

    @property
    def heads(self):
        from dparser.BeamStackBufferState import ID
        heads = [None for _ in self.sentence[0]]
        labels = [None for _ in self.sentence[0]]
        labeled = False
        for parse in self.predecessors:
            if parse.last_move == RIGHT:
                if heads[parse.previous.n0[ID]] is None:
                    heads[parse.previous.n0[ID]] = parse.previous.s0[ID]
                    labeled |= parse.set_label(labels)
            elif parse.last_move == LEFT:
                heads[parse.previous.s0[ID]] = parse.previous.n0[ID]
                labeled |= parse.set_label(labels)
            elif parse.last_move == REDUCE:
                heads[parse.previous.s0[ID]] = parse.s0[ID]
                labeled |= parse.set_label(labels)
        root_first = self.sentence[-2]
        if not root_first:
            heads[-1] = 0
        if labeled: return heads, labels
        return heads

class BeamParser(_BeamParser):
    def _defaultBeamParse(self):
        return BeamParse()

class LocalParser(TransitionSystem, _LocalParser):
    def _valid_moves_elements(self, n0, n, stack, heads):
        root_first = self.root_first
        has_buffer = lambda: n0 <= n
        has_stack = lambda: stack and True
        has_stacks = lambda: len(stack) >= 2
        return root_first, has_buffer, has_stack, has_stacks

    def transition(self, move, n0, stack, parse):
        if stack: s0 = stack[-1]
        monotonic = super().transition(move, n0, stack, parse)
        if move == LEFT:
            # amend RIGHT
            if stack and parse.rights[stack[-1]] and parse.rights[stack[-1]][-1] == s0:
                parse.rights[stack[-1]].pop()
        elif move == REDUCE:
            # amend SHIFT
            if not parse.rights[stack[-1]] or parse.rights[stack[-1]][-1] != s0:
                parse.add(stack[-1], s0)
        return monotonic
